import requests, base64, datetime, types, time, json, pickle, os

from gwa_spotify_api.exceptions import (
    SpotifyTimeoutError, SpotifyAuthenticationError,
    SpotifyNotFoundError, SpotifyInvalidRequestError,
    SpotifyException,
)

MY_SPOTIFY_ID = os.environ.get('MY_SPOTIFY_ID')


def handle_http_errors(f):
    def wrapped_f(*args, **kwargs):
        for i in range(5):
            try:
                response = f(*args, **kwargs)
                return response
            except SpotifyTimeoutError as e:
                retry_after = e.wait()
                time.sleep(retry_after)
            except SpotifyAuthenticationError as e:
                pass
            except (SpotifyNotFoundError, SpotifyInvalidRequestError) as e:
                raise e
        raise SpotifyException(response.json())
    return wrapped_f


class SpotifyAPI(object):
    """API for making request for public data from spotify api ("Client Flow" Authentication style)

    Note: In to get private data, need to use SpotifyAuthAPI class.

    Must have the following environment variables set:

        SPOTIFY_CLIENT_ID
        SPOTIFY_CLIENT_SECRET

    Or you can pass a dictionary with those keys.
    """
    def __init__(self, assign_token=True, config=None, use_default=False):

        self.request_epoch = time.time()
        self.time_between_requests = 1
        self.most_recent_request_address = None

        if config is not None:
            self.config = config
        else:
            self.config = {}
            self.config['SPOTIFY_CLIENT_ID'] = os.environ.get('SPOTIFY_CLIENT_ID')
            self.config['SPOTIFY_CLIENT_SECRET'] = os.environ.get('SPOTIFY_CLIENT_SECRET')

        if assign_token:
            self.assign_token()

    def _get_authorization_string(self, client_id, client_secret):
        encoded_auth_string = (client_id + ':' + client_secret).encode('ascii')
        encoded_auth_string = base64.b64encode(encoded_auth_string)

        return encoded_auth_string.decode('ascii')


    def assign_token(self):
        client_id = self.config['SPOTIFY_CLIENT_ID']
        client_secret = self.config['SPOTIFY_CLIENT_SECRET']

        auth_string = self._get_authorization_string(client_id, client_secret)

        token = requests.post(
            'https://accounts.spotify.com/api/token',
            data={'grant_type':'client_credentials'},
            headers={'Authorization':'Basic ' + auth_string}
        )

        self.token = token.json()['access_token']

    def _get_url_endpoint(self, url, base_url='https://api.spotify.com/v1/'):
        return url[len(base_url):]

    def _get_header(self, header_data=None):
        if hasattr(self, 'token'):
            auth_header = {'Authorization': 'Bearer ' + self.token}
        else:
            auth_header = {}
        if header_data is not None:
            auth_header.update(header_data)
        return auth_header

    def _get_url(self, endpoint, version='v1'):
        return u'https://api.spotify.com/{version}/{endpoint}'.format(
            version=version,
            endpoint=endpoint
        )

    def _handle_response(self, response):
        if response.status_code < 400:
            response = response.json()
            if 'limit' in response.keys():
                return self._page_through_response(response)
            else:
                return response
        if response.status_code == 400:
            raise SpotifyInvalidRequestError(response)
        if response.status_code == 401:
            self.assign_token()
            raise SpotifyAuthenticationError(response)
        if response.status_code == 404:
            raise SpotifyNotFoundError(response)
        if response.status_code == 429:
            raise SpotifyTimeoutError(response)
        if response.status_code >= 500:
            raise SpotifyInternalServerError(response)
        raise SpotifyException(response)

    def _page_through_response(self, response, debug=True):
        items = []
        items.extend(response['items'])
        while True:
            next_page = response['next']
            if next_page is None: 
                break
            next_page = self._get_url_endpoint(next_page)   
            response = self.get(next_page)
            if isinstance(response, dict):
                items.extend(response['items'])
            elif isinstance(response, list):
                items.extend(response)
                break
            else:
                print(response)
                raise Exception("Error while paging through response.  most recent page was not of type dict or list, or if dict it was missing an 'items' field")
        return items

    def _manage_ratelimit(self):
        now = time.time()
        if now - self.request_epoch < self.time_between_requests:
            time.sleep(self.time_between_requests)
        self.request_epoch = now

    def get_most_recent_request_address(self):
        return self.most_recent_request_address

    @handle_http_errors
    def get(self, endpoint, params=None):
        url = self._get_url(endpoint)
        print(u"getting from {0}".format( url))
        self._manage_ratelimit()
        self.most_recent_request_address = endpoint
        r = requests.get(
            url,
            headers=self._get_header(),
            params=params)
        return self._handle_response(r)

    @handle_http_errors
    def post(self, endpoint, params=None, data=None):
        url = self._get_url(endpoint)
        print("posting to {}".format(url))
        self._manage_ratelimit()
        self.most_recent_request_address = endpoint
        r = requests.post(
            url,
            headers=self._get_header(),
            params=params,
            data=json.dumps(data))
        return self._handle_response(r)


    @staticmethod
    def parse_datestring(spotify_date):
        return datetime.datetime.strptime(spotify_date, "%Y-%m-%dT%H:%M:%SZ")

if __name__ == '__main__':

    # test Client Auth Flow
    api = SpotifyAPI()
    print(api.get('users/{}'.format(MY_SPOTIFY_ID)))




import os
import time


ERROR_LOG_FILENAME = 'apierrors.log'

INTERNAL_SERVER_ERROR_WAIT_TIME = 60


class SpotifyException(Exception):

    def __init__(self, response):
        try:
            response_json = response.json()

        except ValueError:
            response_json = response

        print('response: {}'.format(response_json))
        self.response_json = response_json

    def get_response(self):
        return self.response_json['error']

    def get_status_code(self):
        return str(self.response_json['error']['status'])

    def get_message(self):
        return self.response_json['error']['message']

    def log_to_file(self, api, filename=ERROR_LOG_FILENAME):
        most_recent_request = api.get_most_recent_request_address()
        line = '|'.join([self.get_status_code(), self.get_message(), most_recent_request]) + '\n'
        with open(filename, 'a') as logfile:
            logfile.write(line.encode("UTF-8"))


class SpotifyTimeoutError(SpotifyException):

    def __init__(self, response):
        SpotifyException.__init__(self, response)
        self.retry_after = response['headers']['Retry-After']

    def wait(self):
        print("exceeded rate limit.  Trying again in {} seconds...".format(self.retry_after))
        time.sleep(self.retry_after)


class SpotifyAuthenticationError(SpotifyException):

    def __init__(self, response):
        print("token timed out.  Getting new authentication token")

        from gwa_spotify_api.spotify_auth_api import TOKEN_PICKLE_FILENAME

        if TOKEN_PICKLE_FILENAME in os.listdir('.'):
            os.remove(TOKEN_PICKLE_FILENAME )
        SpotifyException.__init__(self, response)

    def authenticate(self):
        return SpotifyAuthAPI()


class SpotifyNotFoundError(SpotifyException):
    pass


class SpotifyInvalidRequestError(SpotifyException):
    pass


class SpotifyInternalServerError(SpotifyException):
    def __init__(self, response):
        SpotifyException.__init__(self, response)
        print("iternal server error.  Going to sleep for {}".format(INTERNAL_SERVER_ERROR_WAIT_TIME))
        time.sleep(wait_time)



import json
import os
import pickle

from rauth import OAuth2Service

from gwa_spotify_api.spotify_api import SpotifyAPI


TOKEN_PICKLE_FILENAME = 'spotify-token.pickle'

SPOTIFY_AUTHORIZE_URL = 'https://accounts.spotify.com/authorize'
SPOTIFY_ACCESS_TOKEN_URL = 'https://accounts.spotify.com/api/token'
SPOTIFY_API_BASE_URL = 'https://accounts.spotify.com'

class SpotifyAuthAPI(SpotifyAPI):

    def __init__(self, assign_token=True, config=None, scopes_list=None):
        SpotifyAPI.__init__(self, assign_token=False, config=config)

        self.callback_url = self.config.get(
            'SPOTIFY_CALLBACK_URL',
            os.environ.get('SPOTIFY_CALLBACK_URL')
        )

        self.client_id = self.config.get(
            'SPOTIFY_CLIENT_ID',
            os.environ.get('SPOTIFY_CLIENT_ID')
        )

        self.client_secret = self.config.get(
            'SPOTIFY_CLIENT_SECRET',
            os.environ.get('SPOTIFY_CLIENT_SECRET')
        )

        if scopes_list is None:
            scopes_list = [
                'user-top-read',
                'user-library-read',
                'user-read-birthdate',
                'playlist-read-private',
                'user-read-private',
                'user-read-email',
                'playlist-modify-public',
                'playlist-modify-private',
                'user-follow-read'
            ]

        self.scope = ' '.join(scopes_list)

        self.service = OAuth2Service(
            name='spotify',
            client_id=self.client_id,
            client_secret=self.client_secret,
            authorize_url=SPOTIFY_AUTHORIZE_URL,
            access_token_url=SPOTIFY_ACCESS_TOKEN_URL,
            base_url=SPOTIFY_API_BASE_URL)

        if assign_token:
           self.assign_token()

    def assign_token(self, token=None):
        if token is not None:
            self.token = token
        else:
            try:
                self.token = pickle.load(open(TOKEN_PICKLE_FILENAME,'rb'))
            except:
                self.get_token_flow()

    def get_authorize_url(self):
        params = dict(
            scope=self.scope,
            response_type='code',
            redirect_uri=self.callback_url)

        url = self.service.get_authorize_url(**params)

        return url

    def get_access_token(self, auth_code):
        token = self.service.get_access_token(
            data={
                'code':auth_code,
                'grant_type':'authorization_code',
                'redirect_uri':callback_url
            },
            method='POST',
            headers={'Authorization': 'Basic ' + self._get_authorization_string(client_id, client_secret)},
            decoder=json.loads)

        self.token = token

        return token


    def get_token_flow(self):
        '''use from terminal'''
        authorize_url = self.get_authorize_url()

        message = (
            '\nvisit the following url and enter the "code" parameter from the url you are redirected to here.'
            '\n\n{}\n'
            '\n--->'
        )
        message = message.format(authorize_url)

        auth_code = input(message)

        token = self.get_access_token(self, auth_code)

        pickle.dump(token, open(TOKEN_PICKLE_FILENAME,'wb'))

if __name__ == '__main__':
    # test Auth API
    api = SpotifyAuthAPI()
    api.assign_token()
    print(api.get('me'))



from os import path
from setuptools import setup, find_packages
import codecs

here = path.abspath(path.dirname(__file__))

with codecs.open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

with open(path.join(here, 'spotify-token.pickle'), 'w') as f:
    f.write('sample-spotify-token')

setup(
    name='gwa_spotify_api',
    version='0.0.6',
    description='wrapper for spotify web api',
    long_description=long_description,
    url='https://github.com/goodwordalchemy/spotify-api',
    author='David Goldberg',
    author_email='goodwordalchemy@gmail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3'
    ],
    keywords='spotify api wrapper',
    packages=find_packages(),
    data_files=[('data', ['spotify-token.pickle'])],
    python_requires='>=3',
    install_requires=['rauth']
)

